import { Store } from 'redux';

const CLEAR_SUGGESTED_ACTIONS = 'WEB_CHAT/CLEAR_SUGGESTED_ACTIONS';

function clearSuggestedActions() {
  return {
    type: CLEAR_SUGGESTED_ACTIONS
  };
}

const CONNECT = 'DIRECT_LINE/CONNECT';

function connect({ directLine, userID, username }) {
  return {
    type: CONNECT,
    payload: {
      directLine,
      userID,
      username
    }
  };
}

const DISCONNECT = 'DIRECT_LINE/DISCONNECT';

function disconnect() {
  return {
    type: DISCONNECT
  };
}

declare const DISMISS_NOTIFICATION = "WEB_CHAT/DISMISS_NOTIFICATION";
type DismissNotificationAction = {
    payload: {
        id: string;
    };
    type: typeof DISMISS_NOTIFICATION;
};
declare function dismissNotification(id: string): DismissNotificationAction;

const EMIT_TYPING_INDICATOR = 'WEB_CHAT/EMIT_TYPING_INDICATOR';

function emitTypingIndicator() {
  return {
    type: EMIT_TYPING_INDICATOR
  };
}

type MarkActivityActionType = 'WEB_CHAT/MARK_ACTIVITY';
type MarkActivityAction = {
    payload: {
        activityID: string;
        name: string;
        value: any;
    };
    type: MarkActivityActionType;
};
declare function markActivity({ id: activityID }: {
    id: string;
}, name: string, value: any): MarkActivityAction;

type AnyAnd<T> = Omit<any, keyof T> & T;

type DirectLineAttachment = {
    content?: any;
    contentType: string;
    contentUrl?: string;
    name?: string;
    thumbnailUrl?: string;
};

type CardActionWithImageAndTitle = {
    image: string;
} | {
    title: string;
} | {
    image: string;
    title: string;
};
/**
 * A `call` action represents a telephone number that may be called.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#call
 */
type CallCardAction = CardActionWithImageAndTitle & {
    type: 'call';
    value: string;
};
/**
 * A `downloadFile` action represents a hyperlink to be downloaded.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#download-file-actions
 */
type DownloadFileCardAction = CardActionWithImageAndTitle & {
    type: 'downloadFile';
    value: string;
};
/**
 * An `imBack` action represents a text response that is added to the chat feed.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#im-back
 */
type IMBackCardAction = CardActionWithImageAndTitle & {
    type: 'imBack';
    value: string;
};
/**
 * A `messageBack` action represents a text response to be sent via the chat system.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#message-back
 */
type MessageBackCardAction = CardActionWithImageAndTitle & {
    displayText?: string;
    text?: string;
    type: 'messageBack';
    value?: {
        [key: string]: any;
    };
};
/**
 * An `openUrl` action represents a hyperlink to be handled by the client.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#open-url-actions
 */
type OpenURLCardAction = CardActionWithImageAndTitle & {
    type: 'openUrl';
    value: string;
};
/**
 * A `playAudio` action represents audio media that may be played.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#play-audio
 */
type PlayAudioCardAction = CardActionWithImageAndTitle & {
    type: 'playAudio';
    value: string;
};
/**
 * A `playVideo` action represents video media that may be played.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#play-video
 */
type PlayVideoCardAction = CardActionWithImageAndTitle & {
    type: 'playVideo';
    value: string;
};
/**
 * A `postBack` action represents a text response that is not added to the chat feed.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#post-back
 */
type PostBackCardAction = CardActionWithImageAndTitle & {
    type: 'postBack';
    value: any;
};
/**
 * A `showImage` action represents an image that may be displayed.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#show-image-file-actions
 */
type ShowImageCardAction = CardActionWithImageAndTitle & {
    type: 'showImage';
    value: string;
};
/**
 * A `signin` action represents a hyperlink to be handled by the client's signin system.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#signin
 */
type SignInCardAction = CardActionWithImageAndTitle & {
    type: 'signin';
    value: string;
};
/**
 * A card action represents a clickable or interactive button for use within cards or as suggested actions. They are used to solicit input from users. Despite their name, card actions are not limited to use solely on cards.
 *
 * https://github.com/Microsoft/botframework-sdk/blob/main/specs/botframework-activity/botframework-activity.md#card-action
 */
type DirectLineCardAction = CallCardAction | DownloadFileCardAction | IMBackCardAction | MessageBackCardAction | OpenURLCardAction | PlayAudioCardAction | PlayVideoCardAction | PostBackCardAction | ShowImageCardAction | SignInCardAction;

type DirectLineSuggestedAction = {
    actions?: DirectLineCardAction[];
    to?: string[];
};

type SupportedRole = 'bot' | 'channel' | 'user';
type SupportedSendStatus = 'sending' | 'send failed' | 'sent';
type ChannelData<SendStatus extends SupportedSendStatus | undefined, Type extends string> = AnyAnd<{
    attachmentSizes?: number[];
    clientActivityID?: string;
    'webchat:sequence-id': number;
    webChat?: {
        /**
         * Time when the activity appear in chat history.
         *
         * Note: if the activity is updated, this value will also be updated.
         */
        receivedAt?: number;
        /**
         * Per-activity style options.
         *
         * New in 4.18.0.
         */
        styleOptions?: Record<string, boolean | number | null | string>;
    };
} & (SendStatus extends SupportedSendStatus ? {
    /**
     * @deprecated Since 4.15.3: Please use `channelData['webchat:send-status']` or `useSendStatusByActivityKey()` hook instead.
     *             Please refer to https://github.com/microsoft/BotFramework-WebChat/pull/4362 for details. This field will be removed on or after 2024-07-31.
     */
    state?: SendStatus;
    /**
     * The send status of the activity.
     *
     * - `"sending"`, the activity is in-transit and it has not been timed out;
     * - `"send failed"`, the activity cannot be delivered permanently and further processing had been stopped;
     * - `"sent"`, the activity is delivered successfully.
     *
     * Due to network-related race conditions, the activity could be marked as `"send failed"` but delivered by the service.
     * In this case, the activity should continue to mark as `"send failed"`.
     *
     * For further details, please see [#4362](https://github.com/microsoft/BotFramework-WebChat/pull/4362).
     */
    'webchat:send-status': SendStatus;
} : {}) & (Type extends 'message' ? {
    messageBack?: {
        displayText: string;
    };
    postBack?: boolean;
    speak?: boolean;
    speechSynthesisUtterance?: SpeechSynthesisUtterance;
    'webchat:fallback-text'?: string;
} : {})>;
type ClientCapabilitiesEntity = {
    requiresBotState?: boolean;
    supportsListening?: boolean;
    supportsTts?: boolean;
    type: 'ClientCapabilities';
};
type Entity = ClientCapabilitiesEntity | {
    type: string;
};
type ChannelAcount<Role extends SupportedRole> = {
    id: string;
    name?: string;
    role: Role;
};
type EventActivityEssence = {
    name: string;
    type: 'event';
    value?: any;
};
type MessageActivityEssence = {
    attachmentLayout?: 'carousel' | 'stacked';
    attachments?: DirectLineAttachment[];
    channelData: {
        streamId?: string;
        streamSequence?: number;
        streamType?: 'final';
    };
    inputHint?: 'accepting' | 'expecting' | 'ignoring';
    locale?: string;
    speak?: string;
    suggestedActions?: DirectLineSuggestedAction;
    text: string | undefined;
    textFormat?: 'markdown' | 'plain' | 'xml';
    type: 'message';
    value?: any;
};
type TypingActivityEssence = {
    type: 'typing';
} | {
    channelData: {
        streamId: string;
        streamSequence: number;
        streamType: 'informative' | 'streaming';
    };
    text: string;
    type: 'typing';
};
type TimestampFromServerEssence = {
    id: string;
    localTimestamp?: string;
    timestamp: string;
};
type TimestampInTransitEssence = {
    id?: string;
    localTimestamp: string;
    timestamp?: string;
};
type TimestampEssence<Role extends SupportedRole, SendStatus extends SupportedSendStatus | undefined> = Role extends 'user' ? SendStatus extends 'sending' | 'send failed' ? TimestampInTransitEssence : TimestampFromServerEssence : TimestampFromServerEssence;
type CoreActivityEssence<Role extends SupportedRole, SendStatus extends SupportedSendStatus | undefined, Type extends string = 'conversationUpdate' | 'event' | 'invoke' | 'message' | 'typing'> = {
    channelData: ChannelData<SendStatus, Type>;
    channelId?: string;
    entities?: Entity[];
    from: ChannelAcount<Role>;
    localTimezone?: string;
    replyToId?: string;
    type: string;
} & TimestampEssence<Role, SendStatus> & (Type extends 'event' ? EventActivityEssence : Type extends 'message' ? MessageActivityEssence : Type extends 'typing' ? TypingActivityEssence : {
    type: Type;
});
type OthersActivity = CoreActivityEssence<'bot' | 'channel', undefined>;
type SelfActivitySendFailed = CoreActivityEssence<'user', 'send failed'>;
type SelfActivitySending = CoreActivityEssence<'user', 'sending'>;
type SelfActivitySent = CoreActivityEssence<'user', 'sent'>;
type SelfActivity = SelfActivitySendFailed | SelfActivitySending | SelfActivitySent;
type WebChatActivity = SelfActivity | OthersActivity;

type PostActivityActionType = 'DIRECT_LINE/POST_ACTIVITY';
type PostActivityAction = {
    meta: {
        method: string;
    };
    payload: {
        activity: WebChatActivity;
    };
    type: PostActivityActionType;
};
declare function postActivity(activity: WebChatActivity, method?: string): PostActivityAction;

const SEND_EVENT = 'WEB_CHAT/SEND_EVENT';

function sendEvent(name, value) {
  return {
    type: SEND_EVENT,
    payload: { name, value }
  };
}

declare function sendFiles(files: readonly Readonly<{
    name: string;
    size: number;
    url: string;
    thumbnail?: string;
}>[]): {
    type: string;
    payload: {
        files: readonly Readonly<{
            name: string;
            size: number;
            url: string;
            thumbnail?: string;
        }>[];
    };
};

type SendBoxAttachment = Readonly<{
    blob: Blob | File;
    thumbnailURL?: URL;
}>;

declare function sendMessage(text: string | undefined, method: string | undefined, { attachments, channelData }?: {
    attachments?: readonly SendBoxAttachment[];
    channelData?: any;
}): {
    type: string;
    payload: {
        attachments: readonly Readonly<{
            blob: Blob | File;
            thumbnailURL?: URL;
        }>[];
        channelData: any;
        method: string;
        text: string;
    };
};

const SEND_MESSAGE_BACK = 'WEB_CHAT/SEND_MESSAGE_BACK';

function sendMessageback(value, text, displayText) {
  return {
    type: SEND_MESSAGE_BACK,
    payload: { displayText, text, value }
  };
}

const SEND_POST_BACK = 'WEB_CHAT/SEND_POST_BACK';

function sendPostback(value) {
  return {
    type: SEND_POST_BACK,
    payload: { value }
  };
}

const SET_DICTATE_INTERIMS = 'WEB_CHAT/SET_DICTATE_INTERIMS';

function setDictateInterims(dictateInterims) {
  return {
    type: SET_DICTATE_INTERIMS,
    payload: { dictateInterims }
  };
}

const SET_DICTATE_STATE = 'WEB_CHAT/SET_DICTATE_STATE';

function setDictateState(dictateState) {
  return {
    type: SET_DICTATE_STATE,
    payload: { dictateState }
  };
}

const SET_LANGUAGE = 'WEB_CHAT/SET_LANGUAGE';

function setLanguage(language) {
  return {
    type: SET_LANGUAGE,
    payload: { language }
  };
}

type Notification = {
    alt: string;
    data: any;
    id: string;
    level: string;
    message: string;
    timestamp: number;
};

declare const SET_NOTIFICATION = "WEB_CHAT/SET_NOTIFICATION";
type SetNotificationAction = {
    payload: Omit<Notification, 'timestamp'>;
    type: typeof SET_NOTIFICATION;
};
declare function setNotification({ alt, data, id, level, message }: Omit<Notification, 'timestamp'>): SetNotificationAction;

const SET_SEND_BOX = 'WEB_CHAT/SET_SEND_BOX';

function setSendBox(text) {
  return {
    type: SET_SEND_BOX,
    payload: { text }
  };
}

declare function setSendBoxAttachments(attachments: readonly SendBoxAttachment[]): {
    type: string;
    payload: {
        attachments: readonly Readonly<{
            blob: Blob | File;
            thumbnailURL?: URL;
        }>[];
    };
};

const SET_SEND_TIMEOUT = 'WEB_CHAT/SET_SEND_TIMEOUT';

function setSendTimeout(sendTimeout) {
  return {
    type: SET_SEND_TIMEOUT,
    payload: { sendTimeout }
  };
}

declare const SET_SEND_TYPING_INDICATOR = "WEB_CHAT/SET_SEND_TYPING_INDICATOR";
type SetSendTypingIndicatorAction = {
    payload: {
        sendTypingIndicator: boolean;
    };
    type: typeof SET_SEND_TYPING_INDICATOR;
};
declare function setSendTypingIndicator(value: any): SetSendTypingIndicatorAction;

const START_DICTATE = 'WEB_CHAT/START_DICTATE';

function startDictate() {
  return { type: START_DICTATE };
}

const START_SPEAKING_ACTIVITY = 'WEB_CHAT/START_SPEAKING';

function startSpeakingActivity() {
  return {
    type: START_SPEAKING_ACTIVITY
  };
}

const STOP_DICTATE = 'WEB_CHAT/STOP_DICTATE';

function stopDictate() {
  return { type: STOP_DICTATE };
}

const STOP_SPEAKING_ACTIVITY = 'WEB_CHAT/STOP_SPEAKING';

function stopSpeakingActivity() {
  return {
    type: STOP_SPEAKING_ACTIVITY
  };
}

const SUBMIT_SEND_BOX = 'WEB_CHAT/SUBMIT_SEND_BOX';

function submitSendBox(method = 'keyboard', { channelData } = {}) {
  return {
    type: SUBMIT_SEND_BOX,
    payload: { channelData, method }
  };
}

type GlobalScopeClock = {
    cancelAnimationFrame?: typeof cancelAnimationFrame;
    cancelIdleCallback?: typeof cancelIdleCallback;
    clearImmediate?: typeof clearImmediate;
    clearInterval: typeof clearInterval;
    clearTimeout: typeof clearTimeout;
    Date: typeof Date;
    requestAnimationFrame?: typeof requestAnimationFrame;
    requestIdleCallback?: typeof requestIdleCallback;
    setImmediate?: typeof setImmediate;
    setInterval: typeof setInterval;
    setTimeout: typeof setTimeout;
};

/**
 * Ponyfills that are supported by Web Chat.
 *
 * If ponyfill is passed, Web Chat will use this ponyfill rather than the one from `globalThis` or `window` object.
 */
type GlobalScopePonyfill = GlobalScopeClock;

type CreateStoreOptions = {
    /**
     * True, to enable Redux development tools, otherwise, false.
     *
     * Because Web Chat use sagas for business logics, some store state are keep at JavaScript heaps.
     * Time-travelling and saving/restoring store state are not supported.
     */
    devTools?: boolean;
    /**
     * Ponyfill to overrides specific global scope members.
     *
     * This option is for development use only. Not all features in Web Chat are ponyfilled.
     *
     * To fake timers, `setTimeout` and related functions can be passed to overrides the original global scope members.
     *
     * Please see [#4662](https://github.com/microsoft/BotFramework-WebChat/pull/4662) for details.
     */
    ponyfill?: Partial<GlobalScopePonyfill>;
};
/**
 * Creates a Redux store internally used by Web Chat.
 *
 * This store is critical for Web Chat business logics to operate, please use with cautions.
 */
declare function withOptions(options: CreateStoreOptions, initialState?: any, ...middlewares: any[]): Store;
/**
 * Creates a Redux store internally used by Web Chat.
 *
 * This store is critical for Web Chat business logics to operate, please use with cautions.
 */
declare function createStore(initialState?: any, ...middlewares: any[]): Store;
/**
 * Creates a Redux store internally used by Web Chat, with Redux development tools.
 *
 * This store is critical for Web Chat business logics to operate, please use with cautions.
 *
 * @deprecated Use `withOptions` instead and pass `{ devTools: true }`
 */
declare function withDevTools(initialState?: any, ...middlewares: any[]): Store;

type OneOrMany<T> = T | T[];

/**
 * The most generic type of item.
 *
 * This is partial implementation of https://schema.org/Thing.
 *
 * @see https://schema.org/Thing
 */
type Thing = {
    '@context'?: 'https://schema.org' | undefined;
    '@id'?: string | undefined;
    '@type': string;
    /**
     * An additional type for the item, typically used for adding more specific types from external vocabularies in microdata syntax. This is a relationship between something and a class that the thing is in. Typically the value is a URI-identified RDF class, and in this case corresponds to the use of rdf:type in RDF. Text values can be used sparingly, for cases where useful information can be added without their being an appropriate schema to reference. In the case of text values, the class label should follow the schema.org [style guide](https://schema.org/docs/styleguide.html).
     *
     * @see https://schema.org/additionalType
     */
    additionalType?: string | undefined;
    /**
     * An alias for the item.
     *
     * @see https://schema.org/alternateName
     */
    alternateName?: string | undefined;
    /**
     * A description of the item.
     *
     * @see https://schema.org/description
     */
    description?: string | undefined;
    /**
     * The name of the item.
     *
     * @see https://schema.org/name
     */
    name?: string | undefined;
    /**
     * Indicates a potential Action, which describes an idealized action in which this thing would play an 'object' role.
     *
     * @see https://schema.org/potentialAction
     */
    potentialAction?: readonly Action[] | undefined;
    /**
     * URL of the item.
     *
     * @see https://schema.org/url
     */
    url?: string | undefined;
};
declare const parseThing: (data: unknown) => Thing;

/**
 * An enterprise (potentially individual but typically collaborative), planned to achieve a particular aim. Use properties from [Organization](https://schema.org/Organization), [subOrganization](https://schema.org/subOrganization)/[parentOrganization](https://schema.org/parentOrganization) to indicate project sub-structures.
 *
 * This is partial implementation of https://schema.org/Project.
 *
 * @see https://schema.org/Project
 */
type Project = Thing & {
    /**
     * A slogan or motto associated with the item.
     *
     * @see https://schema.org/slogan
     */
    slogan?: string | undefined;
};
declare const parseProject: (data: unknown) => Project;

/**
 * An action performed by a direct agent and indirect participants upon a direct object. Optionally happens at a location with the help of an inanimate instrument. The execution of the action may produce a result. Specific action sub-type documentation specifies the exact expectation of each argument/role.
 *
 * See also [blog post](http://blog.schema.org/2014/04/announcing-schemaorg-actions.html) and [Actions overview document](https://schema.org/docs/actions.html).
 *
 * This is partial implementation of https://schema.org/Action.
 *
 * @see https://schema.org/Action
 */
type Action = Thing & {
    /**
     * Indicates the current disposition of the Action.
     *
     * @see https://schema.org/actionStatus
     */
    actionStatus?: 'ActiveActionStatus' | 'CompletedActionStatu' | 'FailedActionStatus' | 'PotentialActionStatus' | undefined;
    /**
     * The service provider, service operator, or service performer; the goods producer. Another party (a seller) may offer those services or goods on behalf of the provider. A provider may also serve as the seller. Supersedes [carrier](https://schema.org/carrier).
     *
     * @see https://schema.org/provider
     */
    provider?: Project | undefined;
};
declare const parseAction: (data: unknown) => Action;

/**
 * A word, name, acronym, phrase, etc. with a formal definition. Often used in the context of category or subject classification, glossaries or dictionaries, product or creative work types, etc. Use the name property for the term being defined, use termCode if the term has an alpha-numeric code allocated, use description to provide the definition of the term.
 *
 * This is partial implementation of https://schema.org/DefinedTerm.
 *
 * @see https://schema.org/DefinedTerm
 */
type DefinedTerm = Thing & {
    /**
     * A [DefinedTermSet](https://schema.org/DefinedTermSet) that contains this term.
     *
     * @see https://schema.org/inDefinedTermSet
     */
    inDefinedTermSet?: string | undefined;
    /**
     * A code that identifies this [DefinedTerm](https://schema.org/DefinedTerm) within a [DefinedTermSet](https://schema.org/DefinedTermSet).
     *
     * @see https://schema.org/termCode
     */
    termCode?: string | undefined;
};
declare const parseDefinedTerm: (data: unknown) => DefinedTerm;

/**
 * The most generic kind of creative work, including books, movies, photographs, software programs, etc.
 *
 * This is partial implementation of https://schema.org/CreativeWork.
 *
 * @see https://schema.org/CreativeWork
 */
type CreativeWork = Thing & {
    /**
     * An abstract is a short description that summarizes a [CreativeWork](https://schema.org/CreativeWork).
     *
     * @see https://schema.org/abstract
     */
    abstract?: string | undefined;
    /**
     * The author of this content or rating. Please note that author is special in that HTML 5 provides a special mechanism for indicating authorship via the rel tag. That is equivalent to this and may be used interchangeably.
     *
     * @see https://schema.org/author
     */
    author?: string | undefined;
    /**
     * A citation or reference to another creative work, such as another publication, web page, scholarly article, etc.
     *
     * @see https://schema.org/citation
     */
    citation?: readonly CreativeWork[] | undefined;
    /**
     * Keywords or tags used to describe some item. Multiple textual entries in a keywords list are typically delimited by commas, or by repeating the property.
     *
     * @see https://schema.org/keywords
     */
    keywords?: readonly (DefinedTerm | string)[] | undefined;
    /**
     * A pattern that something has, for example 'polka dot', 'striped', 'Canadian flag'. Values are typically expressed as text, although links to controlled value schemes are also supported.
     *
     * @see https://schema.org/pattern
     */
    pattern?: DefinedTerm | undefined;
    /**
     * The textual content of this CreativeWork.
     *
     * @see https://schema.org/text
     */
    text?: string | undefined;
    /**
     * The schema.org [usageInfo](https://schema.org/usageInfo) property indicates further information about a [CreativeWork](https://schema.org/CreativeWork). This property is applicable both to works that are freely available and to those that require payment or other transactions. It can reference additional information, e.g. community expectations on preferred linking and citation conventions, as well as purchasing details. For something that can be commercially licensed, usageInfo can provide detailed, resource-specific information about licensing options.
     *
     * This property can be used alongside the license property which indicates license(s) applicable to some piece of content. The usageInfo property can provide information about other licensing options, e.g. acquiring commercial usage rights for an image that is also available under non-commercial creative commons licenses.
     */
    usageInfo?: CreativeWork | undefined;
};
declare const parseCreativeWork: (data: unknown) => CreativeWork;

/**
 * A [Claim](https://schema.org/Claim) in Schema.org represents a specific, factually-oriented claim that could be the [itemReviewed](https://schema.org/itemReviewed) in a [ClaimReview](https://schema.org/ClaimReview). The content of a claim can be summarized with the [text](https://schema.org/text) property. Variations on well known claims can have their common identity indicated via [sameAs](https://schema.org/sameAs) links, and summarized with a name. Ideally, a [Claim](https://schema.org/Claim) description includes enough contextual information to minimize the risk of ambiguity or inclarity. In practice, many claims are better understood in the context in which they appear or the interpretations provided by claim reviews.
 *
 * Beyond [ClaimReview](https://schema.org/ClaimReview), the Claim type can be associated with related creative works - for example a [ScholarlyArticle](https://schema.org/ScholarlyArticle) or [Question](https://schema.org/Question) might be about some [Claim](https://schema.org/Claim).
 *
 * This is partial implementation of https://schema.org/Claim.
 *
 * @see https://schema.org/Claim.
 */
type Claim = CreativeWork & {
    /**
     * Indicates an occurrence of a [Claim](https://schema.org/Claim) in some [CreativeWork](https://schema.org/CreativeWork).
     *
     * @see https://schema.org/appearance.
     */
    appearance?: CreativeWork | undefined;
    /**
     * For a [Claim](https://schema.org/Claim) interpreted from [MediaObject](https://schema.org/MediaObject) content sed to indicate a claim contained, implied or refined from the content of a [MediaObject](https://schema.org/MediaObject).
     *
     * @see https://schema.org/claimInterpreter.
     */
    claimInterpreter?: Project | undefined;
    /**
     * The position of an item in a series or sequence of items.
     *
     * @see https://schema.org/position.
     */
    position?: number | string;
};
declare const parseClaim: (data: unknown) => Claim;

/**
 * An action performed by a direct agent and indirect participants upon a direct object. Optionally happens at a location with the help of an inanimate instrument. The execution of the action may produce a result. Specific action sub-type documentation specifies the exact expectation of each argument/role.
 *
 * See also [blog post](http://blog.schema.org/2014/04/announcing-schemaorg-actions.html) and [Actions overview document](https://schema.org/docs/actions.html).
 *
 * This is partial implementation of https://schema.org/Action.
 *
 * @see https://schema.org/Action
 */
type VoteAction = Action & {
    /**
     * A sub property of object. The options subject to this action. Supersedes [option](https://schema.org/option).
     *
     * @see https://schema.org/VoteAction
     */
    actionOption?: string | undefined;
};
declare const parseVoteAction: (data: unknown) => VoteAction;

declare function getOrgSchemaMessage(graph: readonly WebChatActivity['entities'][0][]): CreativeWork | undefined;

declare function isForbiddenPropertyName(propertyName: string): boolean;

declare function onErrorResumeNext<T>(fn: () => T): T | undefined;
declare function onErrorResumeNext<T, U = T>(fn: () => T, defaultValue: U): T | U;

declare function singleToArray<T>(singleOrArray: T | T[]): T[];

declare function warnOnce(message: string): () => void;

type DirectLineActivity = Exclude<any, [] | boolean | Function | number | string>;

type DirectLineMediaCardEssence = {
    aspect?: '4:3' | '16:9';
    autoloop?: boolean;
    autostart?: boolean;
    buttons?: DirectLineCardAction[];
    duration?: string;
    image?: {
        url: string;
    };
    media: {
        profile?: string;
        url: string;
    }[];
    shareable?: boolean;
    subtitle?: string;
    title?: string;
    value?: any;
};

type DirectLineAnimationCard = DirectLineMediaCardEssence & {
    contentType: 'application/vnd.microsoft.card.animation';
};

type DirectLineAudioCard = DirectLineMediaCardEssence & {
    contentType: 'application/vnd.microsoft.card.audio';
};

type DirectLineCardImage = {
    alt: string;
    tap?: DirectLineCardAction;
    url: string;
};

type DirectLineBasicCardEssence = {
    buttons?: DirectLineCardAction[];
    images?: DirectLineCardImage[];
    subtitle?: string;
    tap?: DirectLineCardAction;
    text?: string;
    title?: string;
};

type DirectLineHeroCard = DirectLineBasicCardEssence & {
    contentType: 'application/vnd.microsoft.card.hero';
};

type DirectLineJSBotConnection = any;

type DirectLineOAuthCard = {
    buttons?: DirectLineCardAction[];
    contentType: 'application/vnd.microsoft.card.oauth';
    text?: string;
};

type Fact = {
    key?: string;
    value?: string;
};
type ReceiptItem = {
    image?: DirectLineCardImage;
    price?: string;
    quantity?: string;
    subtitle?: string;
    tap?: DirectLineCardAction;
    text?: string;
    title?: string;
};
type DirectLineReceiptCard = {
    buttons?: DirectLineCardAction[];
    contentType: 'application/vnd.microsoft.card.receipt';
    facts?: Fact[];
    items?: ReceiptItem[];
    tap?: DirectLineCardAction;
    tax?: string;
    title?: string;
    total?: string;
    vat?: string;
};

type DirectLineSignInCard = {
    buttons?: DirectLineCardAction[];
    contentType: 'application/vnd.microsoft.card.signin';
    text?: string;
};

type DirectLineThumbnailCard = DirectLineBasicCardEssence & {
    contentType: 'application/vnd.microsoft.card.thumbnail';
};

type DirectLineVideoCard = DirectLineMediaCardEssence & {
    contentType: 'application/vnd.microsoft.card.video';
};

/** Receives a completion notification */
type CompleteFunction = () => void;
/** Receives the sequence error */
type ErrorFunction = (error: Error) => void;
/** Receives the next value in the sequence */
type NextFunction<T> = (value: T) => void;
type SubscriberFunction<T> = (observer: SubscriptionObserver<T>) => Subscription;
/** An `Observable` represents a sequence of values which may be observed. */
declare class Observable<T> {
    constructor(subscriber: SubscriberFunction<T>);
    /** Subscribes to the sequence with an observer */
    subscribe(observer: Observer<T>): Subscription;
    /** Subscribes to the sequence with callbacks */
    subscribe(onNext: NextFunction<T>, onError?: ErrorFunction, onComplete?: CompleteFunction): Subscription;
}
declare class Subscription {
    /** Cancels the subscription */
    unsubscribe(): void;
    /** A boolean value indicating whether the subscription is closed */
    get closed(): boolean;
}
/**
 * An `Observer` is used to receive data from an `Observable`, and is supplied as an argument to `subscribe`.
 *
 * All methods are optional.
 */
declare class Observer<T> {
    /** Receives the subscription object when `subscribe` is called */
    start(subscription: Subscription): any;
    /** Receives the next value in the sequence */
    next: NextFunction<T>;
    /** Receives the sequence error */
    error: ErrorFunction;
    /** Receives a completion notification */
    complete: CompleteFunction;
}
/** A `SubscriptionObserver` is a normalized `Observer` which wraps the observer object supplied to `subscribe`. */
declare class SubscriptionObserver<T> {
    /** Sends the next value in the sequence */
    next: NextFunction<T>;
    /** Sends the sequence error */
    error: ErrorFunction;
    /** Sends the completion notification */
    complete: CompleteFunction;
    /** A boolean value indicating whether the subscription is closed */
    get closed(): boolean;
}

declare const Constants: {
    ActivityClientState: any;
    DictateState: any;
};
declare const moduleFormat: string;
declare const version: string;
declare const buildInfo: {
    buildTool: string;
    moduleFormat: string;
    version: string;
};

export { Constants, type DirectLineActivity, type DirectLineAnimationCard, type DirectLineAttachment, type DirectLineAudioCard, type DirectLineCardAction, type DirectLineHeroCard, type DirectLineJSBotConnection, type DirectLineOAuthCard, type DirectLineReceiptCard, type DirectLineSignInCard, type DirectLineSuggestedAction, type DirectLineThumbnailCard, type DirectLineVideoCard, type GlobalScopePonyfill, Observable, type OneOrMany, type Action as OrgSchemaAction, type Claim as OrgSchemaClaim, type CreativeWork as OrgSchemaCreativeWork, type DefinedTerm as OrgSchemaDefinedTerm, type Project as OrgSchemaProject, type Thing as OrgSchemaThing, type SendBoxAttachment, type WebChatActivity, buildInfo, clearSuggestedActions, connect, createStore, withDevTools as createStoreWithDevTools, withOptions as createStoreWithOptions, disconnect, dismissNotification, emitTypingIndicator, getOrgSchemaMessage, isForbiddenPropertyName, markActivity, moduleFormat, onErrorResumeNext, parseAction, parseClaim, parseCreativeWork, parseDefinedTerm, parseProject, parseThing, parseVoteAction, postActivity, sendEvent, sendFiles, sendMessage, sendMessageback as sendMessageBack, sendPostback as sendPostBack, setDictateInterims, setDictateState, setLanguage, setNotification, setSendBox, setSendBoxAttachments, setSendTimeout, setSendTypingIndicator, singleToArray, startDictate, startSpeakingActivity, stopDictate, stopSpeakingActivity, submitSendBox, version, warnOnce };
